#!/usr/bin/env python3
# coding: utf-8

import os
import sys
from importlib import import_module
import psutil
from pathlib import Path
import functools
import traceback
import argparse

from daemon import DaemonContext
from daemon.pidfile import TimeoutPIDLockFile

from ifconf import configure_module, config_callback, configure_main

@config_callback
def config(loader):
    loader.add_attr('module', 'ducts.main', help='module_name containing the main module to run command')
    loader.add_attr('service', 'run', help='method_name to run service process')


def start(pidfile, run):
    if pidfile.is_locked():
        raise Exception('Process is already started.')
    current_dir = os.getcwd()
    with DaemonContext(
            pidfile=pidfile
            , stdout = sys.stdout
            , stderr = sys.stderr):
        os.chdir(current_dir)
        run()
    
def stop(pidfile, run):
    if not pidfile.is_locked():
        return
    pid = pidfile.read_pid()
    for i in range(5):
        try:
            p = psutil.Process(pid)
            print('Terminating PID={}...'.format(pid))
            p.terminate()  #or p.kill()
            p.wait(timeout=3)
        except psutil.TimeoutExpired:
            continue
        except psutil.NoSuchProcess:
            break
    for i in range(3):
        try:
            p = psutil.Process(pid)
            print('Killing PID={}...'.format(pid))
            p.kill()
            p.wait(timeout=1)
        except psutil.TimeoutExpired:
            continue
        except psutil.NoSuchProcess:
            break

        
def main(parser = None, config_arg = 'config_{module}.ini', callback_methods = []):
    parser = parser if parser else argparse.ArgumentParser()
    parser.add_argument('module', help='module_name to run command.')
    parser.add_argument('command', help='[start|stop]')
    main_config = configure_main(argparser = parser, config_arg = config_arg, callback_methods = callback_methods)
    module_config = configure_module(config)
    
    run = getattr(import_module('{}.{}'.format(module_config.module, main_config.args.module)), module_config.service)
    command = getattr(import_module(main.__module__), main_config.args.command)
    
    import ducts.common as c
    pidfile = TimeoutPIDLockFile(c.get_pidfile(main_config.args.module, module_config.service))

    command(pidfile, run)
    
    



from ducts.event import EventHandler

class Handler(EventHandler):

    def __init__(self):
        super().__init__()

    def setup(self, handler_spec, manager):
        handler_spec.set_description('Redisサーバへコマンドを送信し、結果を返します。')
        handler_spec.set_as_responsive()
        return handler_spec

    async def handle(self, event):
        return await event.session.redis.execute(event.data[0], *event.data[1:])




# -*- coding: utf-8 -*-

import inspect
import functools
import itertools

import asyncio
import aioredis

from ifconf import configure_module, config_callback, configure_main

import logging
logger = logging.getLogger(__name__)

@config_callback
def config(loader):
    loader.add_attr('redis_uri_main', 'redis://localhost:6379/0?encoding=utf-8', help='redis uri to connect') # or uri 'unix:///path/to/redis/socket?db=1'
    #loader.add_attr('redis_uri_logging', 'redis://localhost:6379/1?encoding=utf-8', help='redis uri to connect') # or uri 'unix:///path/to/redis/socket?db=1'


class RedisClient:

    def __init__(self, loop):
        self.conf = configure_module(config)
        self.loop = loop
        self.conn = None
        self.conn_for_subscription = None

    async def connect(self):
        self.conn = await aioredis.create_redis_pool(self.conf.redis_uri_main, minsize=1, maxsize=1)
        self.conn_for_subscription = await aioredis.create_redis_pool(self.conf.redis_uri_main, minsize=1, maxsize=1)
        #self.conn_for_logging = await aioredis.create_redis_pool(self.conf.redis_uri_logging, minsize=1, maxsize=1)
        #self.main_conn_shared = await aioredis.create_redis_pool(self.conf.redis_uri_main)
        #self.main_conn_exclusive = await aioredis.create_redis_pool(self.conf.redis_uri_main, minsize=1, maxsize=20)
        #self.logging_conn = await aioredis.create_redis_pool(self.conf.redis_uri_logging)
        #logger.notice('CONNECTED|URL=%s'.format(self.main_conn_shared.address))
        logger.notice('CONNECTED|URL=%s', self.conn.address)

    async def connect_for_blocking(self, minsize, maxsize):
        return await aioredis.create_redis_pool(self.conf.redis_uri_main, minsize=minsize, maxsize=maxsize)

    def execute_threadsafe(self, func, *args, **kwargs):
        #raw_result = self.loop.call_soon_threadsafe(functools.partial(self.conn.execute, cmd, *args))
        async def wrap():
            ret = func(*args, **kwargs)
            if inspect.isawaitable(ret):
                return await ret
            else:
                return ret
        coro = func(*args, **kwargs) if inspect.iscoroutinefunction(func) else wrap()
        future = asyncio.run_coroutine_threadsafe(coro, self.loop)
        while True:
            try:
                ret = future.result(10)
            except asyncio.TimeoutError as e:
                logger.warn('REDIS_EXECUTE_THREADSAFE|TIMEOUT|FUTURE=%s', future)
                continue
            except Exception as e:
                logger.exception('REDIS_EXECUTE_THREADSAFE|ERROR=%s', e)
                ret = -1
                break
            else:
                break
        return ret

    #need to await
    def execute(self, cmd, *args):
        return self.conn.execute(cmd, *args, encoding=None)
            
    #need to await
    def execute_str(self, cmd, *args):
        return self.conn.execute(cmd, *args, encoding='UTF-8')
            
    #need to await
    def subscribe(self, key):
        return self.conn_for_subscription.subscribe(key)
        
    #need to await
    def unsubscribe(self, key_or_channel):
        return self.conn_for_subscription.unsubscribe(key_or_channel)
        
    #need to await
    def psubscribe(self, key):
        return self.conn_for_subscription.psubscribe(key)
        
    #need to await
    def punsubscribe(self, key_or_channel):
        return self.conn_for_subscription.punsubscribe(key_or_channel)
        
    async def xadd(self, streamkey, *args, **kwargs):
        return await self.execute('XADD', streamkey, '*', *args, *itertools.chain.from_iterable(kwargs.items()))
        
    async def xadd_and_publish(self, pubkey, streamkey, *args, **kwargs):
        stream_id = await self.execute('XADD', streamkey, '*', *args, *itertools.chain.from_iterable(kwargs.items()))
        return await self.execute('PUBLISH', pubkey, stream_id)

    async def xget_str(self, streamkey, stream_id):
        ret = await self.execute_str('XREVRANGE', streamkey, stream_id, stream_id, 'COUNT', 1)
        return {v[0] : v[1] for v in zip(*[iter(ret[0][1])]*2)} if ret else {}

    async def xlast_str(self, streamkey):
        ret = await self.execute_str('XREVRANGE', streamkey, '+', '-', 'COUNT', 1)
        return {v[0] : v[1] for v in zip(*[iter(ret[0][1])]*2)} if ret else {}
        
    async def psub_and_xrange_str(self, subkey, streamkey):
        ch = (await self.psubscribe(subkey))[0]
        logger.debug('PSUBSCRIBE|CHANNEL=%s', ch)
        last_id = '0'
        async for msg in ch.iter():
            logger.debug('PSUBSCRIBE|CHANNEL=%s|MSG=%s:%s', ch, msg[0], type(msg[1]))
            ret = await self.execute_str('XRANGE', streamkey, last_id+'1', '+')
            logger.debug('XRANGE|STREAM=%s|LAST_ID=%s|ret=%s', streamkey, last_id, ret)
            if ret is None:
                continue
            for result in ret:
                last_id = result[0]
                yield {v[0] : v[1] for v in zip(*[iter(result[1])]*2)} 
        
    async def psub_and_xrange_str_for_each_id(self, subkey, streamkey):
        ch = (await self.psubscribe(subkey))[0]
        logger.debug('PSUBSCRIBE|CHANNEL=%s', ch)
        async for msg in ch.iter():
            logger.debug('PSUBSCRIBE|CHANNEL=%s|MSG=%s:%s', ch, msg[0], msg[1])
            ret = await self.execute_str('XRANGE', streamkey, msg[1], msg[1])
            logger.debug('XRANGE|STREAM=%s|ID=%s|RET=%s', streamkey, msg, ret)
            if ret is None:
                continue
            for result in ret:
                yield {v[0] : v[1] for v in zip(*[iter(result[1])]*2)}
            
    
    async def close(self):
        #to_close = [con for con in (self.conn_for_subscription, self.conn_for_logging, self.conn) if con != None]
        to_close = [con for con in (self.conn_for_subscription, self.conn) if con != None]
        [con.close() for con in to_close]
        await asyncio.gather(*[con.wait_closed() for con in to_close])

    '''
    async def blocking_command(self, coroutine):
        with await self.main_conn_exclusive as conn:
            await coroutine(conn)

    async def blocking_execute(self, cmd, *args):
        with await redis as r:
            await self.conn.execute(cmd, *args)

    
    async def get_and_wait(self, key):
        value = self.conn.get(key)
        if not value:
            ret = await conn.publish('test', key)
    '''
    



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import re
from pathlib import Path
from datetime import datetime, timedelta
from io import StringIO
from io import BytesIO
from collections import namedtuple
from enum import Enum
import hashlib

import importlib
import inspect
import functools

import msgpack
import json

import asyncio
import aiohttp
from aiohttp import web
import aiohttp_remotes
import aiofiles
import async_timeout
from http import HTTPStatus
import mimetypes

from ifconf import configure_module, config_callback

from .context import ServerContext
from .auth import Auth
from .event import Event, EventSession, HandleType

import logging
        
@config_callback
def config(loader):
    loader.add_attr_boolean('behind_nginx', False, help='True if use ducts behind nginx')
    loader.add_attr('httpd_addr', '0.0.0.0', help='inet address to bind')
    loader.add_attr_int('httpd_port', 8080, help='inet port to bind')
    loader.add_attr_path('local_path_static', Path('./htdocs'), help='local path for static root')
    loader.add_attr_path('local_path_libs', Path('./libs'), hidden=True, help='local path for html/javascript files')

    loader.add_attr_path('module_path_libs', Path('./libs'), hidden=True, help='module path for html/javascript files')
    loader.add_attr_list('web_lib_file_extensions', ['.js', '.html', '.css'], hidden=True, help='web library file extensions')
    loader.add_attr_path('module_path_jsonp_template', Path('./assets/jsonp_template.js'), hidden=True, help='')
    loader.add_attr_path('module_path_favicon', Path('./assets/favicon.ico'), hidden=True, help='')
    
    loader.add_attr('root_path_ducts', '/ducts/', help='URL path for all path')
    loader.add_attr('root_path_static', '/static/', hidden=True, help='URL path for static_root')
    loader.add_attr('root_path_favicon', '/favicon.ico', hidden=True, help='')
    loader.add_attr('relative_path_libs_dir', 'libs', hidden=True, help='URL path for html/javascript files')
    loader.add_attr('relative_path_web_service_discovery', 'wsd', hidden=True, help='where web service discovery is located')
    loader.add_attr('relative_path_web_service_main_jsonp', 'main', hidden=True, help='where main web service is located')
    loader.add_attr('relative_path_event_spec', 'events.json', hidden=True, help='where event descriptions json is located')
    loader.add_attr('relative_path_api', 'api/v1/', hidden=True, help='')
    loader.add_attr('relative_path_websocket', 'ws/v1', hidden=True, help='')
    loader.add_attr('api_path_event', 'event', hidden=True, help='')
    loader.add_attr('api_path_event_id', 'event/{event_id}', hidden=True, help='')
    loader.add_attr('jsonp_callback_query', 'callback', hidden=True, help='')
    loader.add_attr('jsonp_callback_default', 'callback', hidden=True, help='')

    
@config_callback('web_service_discovery')
def wsd_config(loader):
    loader.add_attr('websocket_host', '{ws_host}', help='')
    loader.add_attr('websocket_url', '{ws_scheme}://{ws_host}{root_path_websocket}{session_id}', help='')
    loader.add_attr('websocket_url_reconnect', '{ws_scheme}://{ws_host}{root_path_websocket}{session_id}/reconnect', help='')
    loader.add_attr('api_host', '{api_host}', help='')
    loader.add_attr('api_url_root', '{api_scheme}://{api_host}{root_path_api}', help='')
    loader.add_attr('api_url_event', '{api_scheme}://{api_host}{root_path_api}{api_path_event}', help='')

@config_callback('web_socket_server')
def wsserver_config(loader):
    loader.add_attr_int('ping_pong_interval', 3, help='interval to send ping msg after pong msg received')
    loader.add_attr_int('reconnect_timeout', 60, help='wait timeout until reconnect')

    
class HttpdServer(object):

    def __init__(self, loop):
        self.logger = logging.getLogger(__name__).getChild('httpd')
        self.conf = configure_module(config)
        self.context = ServerContext(loop)
        self.root_path_ducts = Path(self.conf.root_path_ducts)
        self.wsd = configure_module(wsd_config)
        self.wsd_dict = self.wsd._asdict()
        with open(self.context.module_path(self.conf.module_path_jsonp_template), 'r') as f:
            self.jsonp_template = f.read()
        with open(self.context.module_path(self.conf.module_path_favicon), 'rb') as f:
            self.favicon = f.read()
        self.wss = WebSocketServer(self.context)
        self.app = web.Application(loop=self.context.loop)
        self.md5 = hashlib.md5()
        self.setup_route()

    def _root_path_for(self, path, is_dir = False):
        return str(self.root_path_ducts.joinpath(path)) + ('/' if is_dir else '')

    def _root_path_for_websocket(self):
        return self._root_path_for(self.conf.relative_path_websocket, True)
        
    def _root_path_for_api(self):
        return self._root_path_for(self.conf.relative_path_api, True)
        
    def _root_path_for_libs(self):
        return self._root_path_for(self.conf.relative_path_libs_dir, True)
        
    def setup_route(self):
        self.app.router.add_static(self.conf.root_path_static, path=str(self.context.resolve_local_path(self.conf.local_path_static)), name='static', show_index=True)
        self.app.router.add_get(self.conf.root_path_favicon, lambda request: web.Response(body=self.favicon, content_type='image/x-icon'))
        for plugin, level in self.context.plugin_paths(self.conf.module_path_libs, self.conf.local_path_libs):
            for lib_file in [p for p in plugin.glob('**/*') if re.match('.*({})'.format('|'.join(self.conf.web_lib_file_extensions)),str(p))]:
                with open(lib_file, 'rb') as f:
                    buf = f.read()
                    self.md5.update(buf)
                    checksum = self.md5.hexdigest()
                    f = functools.partial((lambda src, mt, request : web.Response(body=src, content_type=mt, headers={'ETag':checksum})), buf, mimetypes.guess_type(str(lib_file))[0])
                self.app.router.add_get(self._root_path_for_libs() + lib_file.name, f)
        self.app.router.add_get(self._root_path_for_libs() + self.conf.relative_path_event_spec
                                , lambda request: web.json_response([s._asdict() for s in self.context.event_handler_manager.specs]))
        self.app.router.add_get(self._root_path_for(self.conf.relative_path_web_service_discovery), self.get_web_service_discovery)
        self.app.router.add_get(self._root_path_for(self.conf.relative_path_web_service_main_jsonp), self.get_web_service_main_jsonp)
        self.app.router.add_get(self._root_path_for_websocket()+'{session_id}', self.wss.websocket_handler)
        self.app.router.add_get(self._root_path_for_websocket()+'{session_id}/reconnect', self.wss.websocket_handler_reconnect)
        self.app.router.add_get(self._root_path_for_api()+self.conf.api_path_event, self.handle_event)
        self.app.router.add_get(self._root_path_for_api()+self.conf.api_path_event_id, self.handle_event_id)

    async def run(self):
        if self.conf.behind_nginx:
            await aiohttp_remotes.setup(self.app, aiohttp_remotes.XForwardedRelaxed())
        self.handler = self.app.make_handler()
        self.srv = await self.context.loop.create_server(self.handler, self.conf.httpd_addr, self.conf.httpd_port)
        self.logger.notice('START|SOCKET=%s', self.srv.sockets[0].getsockname())
        await self.context.redis.connect()

    async def close(self):
        self.logger.debug('CLOSEING|REDIS')
        await self.context.close()
        self.logger.debug('CLOSEING|HTTPD')
        self.srv.close()
        await self.srv.wait_closed()
        self.logger.debug('CLOSEING|APP')
        await self.app.shutdown()
        self.logger.debug('CLOSEING|HANDLERS')
        await self.handler.shutdown(60.0)
        self.logger.debug('CLEANING|APP')
        await self.app.cleanup()
        self.logger.debug('CLOSED')

    async def get_web_service_discovery(self, request):
        return web.json_response(await self._create_wsd_dict(request), status=HTTPStatus.OK.value)

    async def get_web_service_main_jsonp(self, request):
        params = request.rel_url.query
        callback = params.get(self.conf.jsonp_callback_query, self.conf.jsonp_callback_default)
        libs = self._root_path_for_libs()
        script = self.jsonp_template.replace('__TEMPLATE_DUCTS_LIBS__', libs).replace('__TEMPLATE_CALLBACK__', callback).replace('__TEMPLATE_WSD__', json.dumps(await self._create_wsd_dict(request)))
        res = web.Response(body=script, content_type='application/javascript')
        res.last_modified = datetime.now()
        return res
        
    async def handle_event(self, request):
        return web.json_response(self.context.event_handler_manager.key_ids)

    async def handle_event_id(self, request):
        operation = request.match_info['event_id']
        params = request.rel_url.query['args'].strip().split()
        return web.json_response({'result' : await self.redis.execute(operation, params)})

    async def _create_wsd_dict(self, request):
        params = self.conf._asdict()
        params['root_path_websocket'] = self._root_path_for_websocket()
        params['ws_host'] = request.host
        params['ws_scheme'] = 'wss' if request.secure else 'ws'
        params['root_path_api'] = self._root_path_for_api()
        params['api_host'] = request.host
        params['api_scheme'] = request.scheme
        params['session_id'] = await self.wss.init_session(request)
        your_wsd = namedtuple('wsd', ' '.join(self.wsd_dict.keys()))(*[v.format(**params) if type(v) is str else v for v in self.wsd_dict.values()])
        your_wsd_dict = your_wsd._asdict()
        your_wsd_dict['EVENT'] = self.context.event_handler_manager.key_ids
        return your_wsd_dict
    
class WebSocketSession():

    def __init__(self, context, sid, reconnect_timeout):
        self.context = context
        self.sid = sid
        self.context.session_store[sid] = self
        self.reconnect_timeout = reconnect_timeout
        self.reconnect_lock = None
        self.reconnect_until = 0
        self.request = None
        self.socket = None
        self.lock_alive_monitoring = None
        self.logger = logging.getLogger(__name__).getChild('ws').getChild(self.sid)

    def is_closed(self):
        return self.socket is self.reconnect_lock is None
    
    async def prepare(self, request, reconnect = False):
        reconnect_interval = max(datetime.now().timestamp() - (self.reconnect_until - self.reconnect_timeout), 0)
        self.reconnect_until = 0
        self.request = request

        if not reconnect:
            self.socket = None
            if self.reconnect_lock:
                lock = self.reconnect_lock
                self.reconnect_lock = None
                async with lock:
                    lock.notify_all()
            while self.lock_alive_monitoring:
                lock = self.lock_alive_monitoring
                try:
                    async with async_timeout.timeout(5) as cm:
                        async with lock:
                            await lock.wait()
                except asyncio.TimeoutError:
                    pass
            self.context.session_store[self.sid] = self

        self.socket = web.WebSocketResponse()
        await self.socket.prepare(self.request)
        asyncio.ensure_future(self.ping())
        
        if reconnect:
            if self.reconnect_lock:
                lock = self.reconnect_lock
                self.reconnect_lock = None
                async with lock:
                    lock.notify_all()

    async def start_alive_monitoring(self):
        self.lock_alive_monitoring = asyncio.Condition()
        
    async def end_alive_monitoring(self):
        if self.lock_alive_monitoring:
            lock = self.lock_alive_monitoring
            self.lock_alive_monitoring = None
            async with lock:
                lock.notify_all()
        self.context.session_store.pop(self.sid)
                
    async def closed_by_client(self):
        logger.debug("CLOSED_BY_CLIENT|SOCKET=%s|TRANSPORT=%s|IS_CLOGING=%s", self.socket.closed, self.request.transport, is_closing)

    async def start(self):
        self.logger.info('START')
        
    async def end(self, msg):
        self.logger.info('MSG=%s|END|TYPE=%s', id(msg), str(msg.type))

    async def send_str(self, data):
        assert self.socket is not None, 'session must be prepared to send_str'
        await self.socket.send_str(data)

    async def ping(self):
        assert self.socket is not None, 'session must be prepared to ping'
        await self.socket.ping()

    async def pong(self):
        assert self.socket is not None, 'session must be prepared to pong'
        await self.socket.pong()

    async def send_bytes_if_available(self, request_id, event_id, data, loop = False, logger = None):
        log = logger if logger else self.logger
        if self.socket is not None and not self.socket.closed:
            try:
                log.debug("SEND_BYTES_AWAIT|DATA=%s", type(data))
                await asyncio.shield(self.socket.send_bytes(msgpack.packb([request_id, event_id, data])))
                log.debug("SEND_BYTES_DONE|NEXT=%s", loop)
                return True
            except Exception as e:
                log.exception("SEND_BYTES_ERROR=%s", e)
                
    async def send_bytes_or_wait_reconnect(self, request_id, event_id, data, loop = False, logger = None):
        log = logger if logger else self.logger
        while self.socket is not None or self.reconnect_lock is not None:
            if self.socket is not None:
                if not self.socket.closed and self.request.transport is not None:
                    is_closing = self.request.transport.is_closing()
                    logger.debug("SEND_BYTES|TRANSPORT=%s|IS_CLOGING=%s", self.request.transport, is_closing)
                    try:
                        log.debug("SEND_BYTES_AWAIT|DATA=%s", type(data))
                        await asyncio.shield(self.socket.send_bytes(msgpack.packb([request_id, event_id, data])))
                        log.debug("SEND_BYTES_DONE|NEXT=%s", loop)
                        return True
                    except Exception as e:
                        log.exception("SEND_BYTES_ERROR=%s", e)
                        try:
                            if self.socket is not None:
                                log.exception("SEND_BYTES_ERROR=%s", e)
                                await self.socket.close()
                        except Exception as e:
                            self.logger.exception("SOCKET_CLOSE_ERROR=%s", e)
                logger.debug("SOCKET_CLOSED|TRANSPORT=%s", self.request.transport)
                self.reconnect_lock = self.reconnect_lock if self.reconnect_lock else asyncio.Condition()
                self.reconnect_until = self.reconnect_until if self.reconnect_until > 0 else datetime.now().timestamp() + self.reconnect_timeout
                self.socket = None
            if self.reconnect_lock is not None:
                lock = self.reconnect_lock
                timeout = self.reconnect_until - datetime.now().timestamp()
                if timeout > 0:
                    try:
                        async with async_timeout.timeout(timeout) as cm:
                            async with lock:
                                await lock.wait()
                    except asyncio.TimeoutError:
                        pass
                else:
                    async with lock:
                        lock.notify_all()
                    self.reconnect_lock = None
        return False
        

class WebSocketEventSession(EventSession):

    def __init__(self, session, msg, request_id, event_id, data):
        super().__init__(session.context.event_handler_manager)
        self.logger = logging.getLogger(__name__).getChild('event').getChild(str(event_id)).getChild(session.sid).getChild(str(id(msg)))
        self.context = session.context
        self._ss = session
        self._socket = session.socket
        self._msg_id = id(msg)
        self._request_id = request_id
        self._event_id = event_id
        self._event_data = data
        
    async def is_closed(self):
        return self._ss.is_closed()

    def request_id(self):
        return self._request_id

    def session_id(self):
        return self._ss.sid

    #need to await
    def log_message(self, request_id, event_id, data):
        return 

    async def send_bytes_if_available(self, request_id, event_id, data, loop = False):
        return await self._ss.send_bytes_if_available(request_id, event_id, data, loop, logger=self.logger)
    
    async def send_bytes_or_wait_reconnect(self, request_id, event_id, data, loop = False):
        return await self._ss.send_bytes_or_wait_reconnect(request_id, event_id, data, loop, logger=self.logger)

    async def send_str(self, data):
        await self._ss.send_str(data)

    async def ping(self):
        await self._ss.ping()

    async def pong(self):
        await self._ss.pong()

    async def closed_by_client(self):
        await self._ss.closed_by_client()

    async def start_alive_monitoring(self):
        self.logger.debug("START|ALIVE_MONITORING")
        await self._ss.start_alive_monitoring()
        
    async def end_alive_monitoring(self, closed):
        self.logger.info("HANDLER_CLOSED=%s", closed)
        await self._ss.end_alive_monitoring()
        
class WebSocketServer(object):
    
    def __init__(self, context):
        self.conf = configure_module(wsserver_config)
        self.context = context
        self.auth = Auth(context)
        self.wsmsg_handler = {msg_type : lambda : self._default_wsmsg_handler
                              for msg_type in [getattr(aiohttp.WSMsgType, attr) for attr in dir(aiohttp.WSMsgType)]
                              if type(msg_type) is aiohttp.WSMsgType}
        self.wsmsg_handler[aiohttp.WSMsgType.BINARY] = lambda : self._msgpack_msg_handler
        #self.wsmsg_handler[aiohttp.WSMsgType.CONTINUATION] = lambda : self._msgpack_msg_handler
        #self.wsmsg_handler[aiohttp.WSMsgType.TEXT] = lambda : self._default_textmsg_handler
        self.wsmsg_handler[aiohttp.WSMsgType.PING] = lambda : self._pingmsg_handler
        self.wsmsg_handler[aiohttp.WSMsgType.PONG] = lambda : self._pongmsg_handler
        self.wsmsg_handler[aiohttp.WSMsgType.CLOSE] = lambda : self._closemsg_handler

        self.event_handler = [None for v in HandleType]
        self.event_handler[HandleType.ALIVE_MONITORING.value] = self._event_handler_alive_monitoring
        self.event_handler[HandleType.ASYNC_FOR.value] = self._event_handler_async_for
        self.event_handler[HandleType.FOR.value] = self._event_handler_for
        self.event_handler[HandleType.ASYNC.value] = self._event_handler_async
        self.event_handler[HandleType.SYNC.value] = self._event_handler_sync

    async def init_session(self, request):
        return await self.auth.init_session(request)

    async def websocket_handler(self, request):
        sid = await self.auth.check_session(request)
        if sid in self.context.session_store:
            session = self.context.session_store[sid]
        else:
            session = WebSocketSession(self.context, sid, self.conf.reconnect_timeout)
        session.logger.debug("OPEN|REQUEST")
        await session.prepare(request)
        return await self.web_socket_session_loop(session)

    async def websocket_handler_reconnect(self, request):
        sid = await self.auth.check_session(request)
        if sid not in self.context.session_store:
            raise web.HTTPForbidden()
        session = self.context.session_store[sid]
        session.logger.debug("RECONNECT|REQUEST")
        await session.prepare(request, reconnect=True)
        return await self.web_socket_session_loop(session)

    async def web_socket_session_loop(self, session):
        await session.start()
        async for msg in session.socket:
            session.logger.debug('MSG=%s|START|TYPE=%s', id(msg), str(msg.type))
            handle = self.wsmsg_handler[msg.type]()
            try:
                assert inspect.iscoroutinefunction(handle), 'handle must be async function'
                asyncio.ensure_future(handle(session, msg))
            except Exception as e:
                session.logger.exception('MSG=%s|HANDLE=%s|ERROR=%s', id(msg), handle, e)
            session.logger.debug('MSG=%s|NEXT', id(msg))
            await session.end(msg)
        return session.socket

    async def _default_wsmsg_handler(self, session, msg):
        session.logger.debug("MSG=%s|UNHNADLED", id(msg))
        await session.send_str('N/A')
        
    async def _pingmsg_handler(self, session, msg):
        session.logger.debug("MSG=%s|PING", id(msg))
        await session.pong()

    async def _pongmsg_handler(self, session, msg):
        session.logger.debug("MSG=%s|PONG", id(msg))
        await asyncio.sleep(self.conf.ping_pong_interval)
        await session.ping()

    async def _closemsg_handler(self, session, msg):
        session.logger.debug("MSG=%s|CLOSE", id(msg))
        await session.closed_by_client()

    async def _msgpack_msg_handler(self, socket_session, msg):
        request_id, event_id, data = msgpack.unpackb(msg.data)
        event_id = int(event_id)
        session = WebSocketEventSession(socket_session, msg, request_id, event_id, data)
        session.logger.debug("REQUEST=%s|TYPE=%s|LEN=%s", request_id, type(data), len(data) if hasattr(data, '__len__') else 0)
        try:
            handle_type, handler = self.context.event_handler_manager.get_handler_for(event_id)
            await self.event_handler[handle_type.value](session, msg, request_id, event_id, data, handler.handle, Event(event_id, session, data))
            session.logger.debug("END".format(session.session_id(), id(msg), event_id))
        except Exception as e:
            session.logger.exception("END|ERROR=%s", e)
            if not await session.is_closed():
                event_id = -1 * event_id
                ret = '{}:{}'.format(e.__class__.__name__, e.args)
                await session.send_bytes_if_available(request_id, event_id, ret)


    async def _event_handler_alive_monitoring(self, session, msg, request_id, event_id, data, handle, event):
        await session.start_alive_monitoring()
        async for ret in handle(event):
            if not await session.send_bytes_or_wait_reconnect(request_id, event_id, ret, True):
                break
        session.logger.debug("START_CLOSE_HANDLERS")
        closed = await asyncio.gather(*[f for f in [h[1].handle_closed(session) for h in session.manager.handlers()] if inspect.isawaitable(f)])
        await session.end_alive_monitoring(closed)
            
    async def _event_handler_async_for(self, session, msg, request_id, event_id, data, handle, event):
        session.logger.debug("START|AWAIT_ASYNC_LOOP")
        async for ret in handle(event):
            if not await session.send_bytes_or_wait_reconnect(request_id, event_id, ret, True):
                return False
        return True

    async def _event_handler_for(self, session, msg, request_id, event_id, data, handle, event):
        session.logger.debug("START|SYNC_LOOP")
        for ret in handle(event):
            if not await session.send_bytes_or_wait_reconnect(request_id, event_id, ret, True):
                return False
        return True

    async def _event_handler_async(self, session, msg, request_id, event_id, data, handle, event):
        session.logger.debug("START|AWAIT_ASYNC_HANDLE")
        ret = await handle(event)
        return await session.send_bytes_or_wait_reconnect(request_id, event_id, ret)

    async def _event_handler_sync(self, session, msg, request_id, event_id, data, handle, event):
        session.logger.debug("START|SYNC_HANDLE")
        ret = handle(event)
        return await session.send_bytes_or_wait_reconnect(request_id, event_id, ret)

        



#!/usr/bin/env python3

import sys
import os.path

import traceback


import threading
from concurrent.futures import ThreadPoolExecutor
#from concurrent.futures import ProcessPoolExecutor

import functools
#from queue import Queue, Empty
from io import BytesIO
import time
from datetime import datetime

import asyncio
import aiofiles

from google.cloud import speech
from google.cloud.speech import enums
from google.cloud.speech import types

from google.cloud import speech_v1
from google.cloud.speech_v1 import enums

from ifconf import configure_module, config_callback, configure_main

from ducts.redis import RedisClient

import logging

        
@config_callback()
def config(loader):
    loader.add_attr('key', 'ASR/QUEUE', hidden=True, help='list key for asr')
    loader.add_attr_int('timeout_list_for_asr', 10, hidden=True, help='timeout for asr waiting in sec. can be 0.')
    loader.add_attr_int('timeout_stream_read', 2000, hidden=True, help='timeout for stream read in msec. cannot be 0.')
    loader.add_attr_int('timeout_thread_asr_lock', 10, hidden=True, help='timeout for asr waiting in sec. can be 0.')
    loader.add_attr('asr_language_code', 'ja-JP', help='')
    loader.add_attr_int('asr_sample_rate_hertz', 16000, help='')
    loader.add_attr_boolean('debug_store_pcm_enabled', False, help='')
    loader.add_attr('debug_store_pcm_fmt', './{}.{}.{}', help='')
    loader.add_attr_float('timeout_no_response', 0.5, hidden=True, help='timeout for wait response in sec. if 0, no timeout.')
    

class AutomaticSpeechRecognizerManager(object):

    def __init__(self, loop):
        self.conf = configure_module(config)
        self.loop = loop
        self.redis = RedisClient(loop)
        self.thread_pool = ThreadPoolExecutor(max_workers=256)
        self.speech_client = speech.SpeechClient()
        
        self.audio_stream_key_dict = {}
        self.audio_input_lock = asyncio.Condition()
        
    async def run(self):
        await self.redis.connect()
        self.reids_connection = await self.redis.connect_for_blocking(2, 8)
        asyncio.ensure_future(self.listen_list_for_asr())
        asyncio.ensure_future(self.listen_audio_input())

    async def listen_list_for_asr(self):
        logger = logging.getLogger(__name__).getChild('manager').getChild('listen_list')
        listkey = self.conf.key
        logger.info('START|KEY=%s', listkey)
        with await self.reids_connection as conn:
            while True:
                try:
                    logger.debug('BLOCK|TIMEOUT=%s', self.conf.timeout_list_for_asr)
                    spec_key = await conn.execute('BLPOP', listkey, self.conf.timeout_list_for_asr, encoding=None)
                    if spec_key:
                        logger.info('POP|SPEC_KEY=%s', spec_key[1])
                        spec = await self.redis.xlast_str(spec_key[1])
                        logger.debug('POP|SPEC=%s', spec)
                        if not spec:
                            logger.warn('INVALID_SPEC_KEY|EMPTY|SPEC_KEY=%s', spec_key[1])
                            continue
                        handler = AutomaticSpeechRecognizerThread(spec, self)
                        ret = await self.redis.execute('XREVRANGE', handler.stream_key, '+', '-', 'COUNT', 1)
                        handler.stream_id = ret[0][0] if ret else handler.stream_id
                        handler.future = self.loop.run_in_executor(self.thread_pool, handler.run)
                        asyncio.ensure_future(handler.future)
                        self.audio_stream_key_dict[handler.stream_key] = handler
                        if len(self.audio_stream_key_dict) == 1:
                            async with self.audio_input_lock:
                                self.audio_input_lock.notify_all()
                except Exception as e:
                    logger.exception('ERROR=%s', e)
        logger.info('END|KEY=%s', listkey)
        
            
    async def listen_audio_input(self):
        logger = logging.getLogger(__name__).getChild('manager').getChild('listen_audio')
        logger.info('START')
        with await self.reids_connection as conn:
            while True:
                try:
                    while not self.audio_stream_key_dict:
                        logger.debug('LOCK_WAIT')
                        async with self.audio_input_lock:
                            await self.audio_input_lock.wait()
                        logger.debug('LOCK_NOTIFIED|KEYS=%s', self.audio_stream_key_dict)
                    keys = list(self.audio_stream_key_dict.keys())
                    values = [h.stream_id for h in self.audio_stream_key_dict.values()]
                    logger.debug('BLOCK|STREAM=%s|TIMEOUT=%s', self.audio_stream_key_dict.items(), self.conf.timeout_stream_read)
                    ret = await conn.execute('XREAD', 'COUNT', 10, 'BLOCK', self.conf.timeout_stream_read, 'STREAMS', *keys, *values, encoding=None)
                    if not ret:
                        for del_key in [k for k,v in self.audio_stream_key_dict.items() if v.flg_exit]:
                            self.audio_stream_key_dict.pop(del_key)
                        continue
                    for stream_result in ret:
                        stream_key = stream_result[0].decode('UTF-8')
                        logger.debug('XREAD|STREAM=%s', stream_key)
                        handler = self.audio_stream_key_dict[stream_key]
                        for result in stream_result[1]:
                            id = result[0]
                            value = {v[0].decode('UTF-8') : v[1] for v in zip(*[iter(result[1])]*2)} 
                            handler.append(id, value)
                except Exception as e:
                    logger.exception('ERROR|TYPE=%s', e)
                    await asyncio.sleep(1)
        logger.info('END')

    async def close(self):
        await self.redis.close()

file_index = 0
def next_index():
    global file_index
    file_index += 1
    return file_index
    
class AutomaticSpeechRecognizerThread(object):

    def __init__(self, spec, manager):
        for key,value in spec.items():
            setattr(self, key, value)
        self.stream_key = self.AUDIO_STREAM_KEY
        self.asr_pubsub_key = self.ASR_PUBSUB_KEY
        self.asr_stream_key = self.ASR_STREAM_KEY
        self.result_pubsub_key = self.RESULT_PUBSUB_KEY
        self.result_stream_key = self.RESULT_STREAM_KEY
        
        self.manager = manager
        self.conf = manager.conf
        self.speech_client = manager.speech_client
        self.stream_id = '0'
        self.future = None
        
        self.queue = []
        self.lock = threading.Condition()
        self.time_last_response = 0
        self.flg_exit = False
        
        self.filename = self.stream_key.split('/')[0].replace(':','.')


    def generator(self):
        self.path_index = next_index()
        thread_id = threading.get_ident()
        logger = logging.getLogger(__name__).getChild('generator').getChild(str(thread_id))
        logger.info('START|KEY=%s', self.stream_key)
        if self.conf.debug_store_pcm_enabled:
            with open(self.conf.debug_store_pcm_fmt.format(self.filename, self.path_index, 'pcm'), 'wb') as f:
                pass
        while not self.flg_exit:
            try:
                logger.debug('LOCK|QUEUE=%s', len(self.queue))
                with self.lock:
                    loop_count = 0
                    while len(self.queue) == 0:
                        if not self.lock.wait(5):
                            logger.debug('WAIT_TIMEOUT|LOOP_COUNT=%s|FLG_EXIT=%s', loop_count, self.flg_exit)
                            loop_count += 1
                            if self.flg_exit:
                                raise GeneratorExit()
                    if None in self.queue:
                        self.flg_exit = True
                        raise GeneratorExit()
                    data = b''.join(self.queue)
                    self.queue.clear()
                logger.debug('NOTIFIED|DATA=%s', len(data))
                if self.conf.debug_store_pcm_enabled:
                    with open(self.conf.debug_store_pcm_fmt.format(self.filename, self.path_index, 'pcm'), 'ab') as f:
                        f.write(data)
                yield data
                waiting = time.time() - self.time_last_response if self.time_last_response > 0 else 0
                if waiting > self.conf.timeout_no_response and self.conf.timeout_no_response > 0:
                    self.time_last_response = 0
                    logger.debug('NO_RESPONSE|WAIT=%s', waiting)
                    break
            except GeneratorExit as e:
                logger.warn('GENERATOR_EXIT')
                #raise e
                break
            except Exception as e:
                logger.exception('ERROR=%s', e)
                break
        logger.info('EXIT')

    def run(self):
        thread_id = threading.get_ident()
        logger = logging.getLogger(__name__).getChild('asr_thread').getChild(str(thread_id))
        logger.info('RUN|KEY=%s', self.stream_key)
        try:
            #loop = asyncio.new_event_loop()
            #asyncio.set_event_loop(loop)

            pubkey = self.asr_pubsub_key + str(datetime.now().timestamp())
            streamkey = self.asr_stream_key
            self.manager.redis.execute_threadsafe(
                self.manager.redis.xadd_and_publish
                , pubkey
                , streamkey
                , STATE = 'RUNNING'
                , REQUEST_ID='ASR'
                , THREAD = thread_id)
            
            with self.lock:
                while len(self.queue) == 0:
                    logger.debug('START_LOCK|QUEUE=%s|TIMEOUT=%s', len(self.queue), self.conf.timeout_thread_asr_lock)
                    ret = self.lock.wait(self.conf.timeout_thread_asr_lock)
            logger.debug('START_LOCK_END|QUEUE=%s', len(self.queue))
                        
            config = types.RecognitionConfig(
                encoding = enums.RecognitionConfig.AudioEncoding.LINEAR16
                , sample_rate_hertz = self.manager.conf.asr_sample_rate_hertz
                , language_code = self.manager.conf.asr_language_code
            )
            streaming_config = types.StreamingRecognitionConfig(
                config = config
                , interim_results = True
                , single_utterance = False
            )
            while not self.flg_exit:
                self.time_last_response = 0
                logger.info('START|KEY=%s', self.stream_key)

                #https://github.com/googleapis/google-cloud-python/issues/6077
                #client = speech.SpeechClient()
                requests = (types.StreamingRecognizeRequest(audio_content=content) for content in self.generator())

                logger.debug('1.STREAM_RECOGNIZE|CONFIG=%s', streaming_config)
                responses = self.speech_client.streaming_recognize(streaming_config, requests)
                logger.debug('2.GOT_RESPONSES=%s', responses)

                for count, response in enumerate(responses):
                    self.time_last_response = time.time()
                    logger.debug('3.1.RESPONSE|%s=%s', count, response)
                    #if utterance != "" and time.time() - prev_get_time >= 0.5: # 最後に発話を取得してから0.5秒後に送信
                    #raise Exception
                    #    break
                    if not (response.results and response.results[0].alternatives):
                        logger.debug('3.x.RESPONSE_WITH_NO_RESULT')
                        continue
                    result = response.results[0]
                    utterance = result.alternatives[0].transcript
                    is_final = result.is_final
                    prob = result.alternatives[0].confidence if is_final else result.stability
                    result_end_time = result.result_end_time.seconds + result.result_end_time.nanos / 1000000000
                    logger.debug('3.2.RESULT%s(%s,%s)|%s', count, is_final, prob, utterance)

                    pubkey = self.result_pubsub_key + str(datetime.now().timestamp())
                    streamkey = self.result_stream_key
                    self.manager.redis.execute_threadsafe(
                        self.manager.redis.xadd_and_publish
                        , pubkey
                        , streamkey
                        , BEHAVIOR=utterance
                        , REQUEST_ID='ASR'
                        , IS_FINAL=1 if is_final else 0
                        , PROB=prob
                        , RESULT_END_TIME=result_end_time)
                    if is_final:
                        break
        except Exception as e:
            logger.exception('ERROR=%s', e)
            state = 'ERROR'
            error = traceback.format_exc()
        else:
            state = 'END'
            error = ''

        pubkey = self.asr_pubsub_key + str(datetime.now().timestamp())
        streamkey = self.asr_stream_key
        self.manager.redis.execute_threadsafe(
            self.manager.redis.xadd_and_publish
            , pubkey
            , streamkey
            , STATE = state
            , TRACEBACK = error
            , REQUEST_ID='ASR'
            , THREAD = thread_id)
        self.flg_exit = True
        logger.info('EXIT')
                
    def append(self, id, value):
        self.stream_id = id
        with self.lock:
            self.queue.append(value.get('AUDIO', None))
            self.lock.notify_all()
        
        



#!/usr/bin/env python3
# coding: utf-8

import sys

from .app import main
    
if __name__ == '__main__':
    import ducts.server
    import ducts.redis
    import ducts.backend.asr

    callback_methods = [ducts.common.config, ducts.redis.config, ducts.server.config, ducts.server.wsserver_config]
    if len(sys.argv) > 1:
        if sys.argv[1] == 'asr':
            callback_methods = [ducts.common.config, ducts.redis.config, ducts.backend.asr.config]

    main(config_arg = 'config_{module}.ini', callback_methods = callback_methods)
    
